#include "CIOboardTCP.h"
#include <iostream>
#include <string.h>

CIOBoardTCP::CIOBoardTCP(const unsigned BoardModel, const std::string Address, const unsigned Port, const bool CreateOwnThread, const bool OnlyMonitoring)
	: CIOBoard(OnlyMonitoring, BoardModel)
	, address(Address)
	, port(Port)
	, stop_thread(false)
	, own_thread(CreateOwnThread)
	, get_is_set(false)
	, settings_get(NULL)
	, settings_size(0)
{
	if(own_thread)
		timer_thread = std::thread(&CIOBoardTCP::TimerThread,this);
	switch(board_model)
	{
		default:
		case CANLAB_BOARD_MODEL_1:
			settings_size = sizeof(CANLAB_IO_BOARD_MODEL_1_SETTINGS);
			break;
		case VRG_BOARD_MODEL_1:
			settings_size = sizeof(VRG_IO_BOARD_MODEL_1_SETTINGS);
			break;
		case VRG_BOARD_MODEL_2:
			settings_size = sizeof(VRG_IO_BOARD_MODEL_2_SETTINGS);
			break;
	}
	settings_get = new unsigned char[settings_size];
}

CIOBoardTCP::~CIOBoardTCP()
{
	SetNonBlocking(true);
	if(own_thread)
	{
		stop_thread = true;
		if(timer_thread.joinable()) 
			timer_thread.join();
	}
	if(is_connected)
	{
	#ifdef _LINUX
        close(socket_id);
	#else
        closesocket(socket_id);
    #endif
	}
	delete[] settings_get;
}
 
bool CIOBoardTCP::Connect(void)
{
	#ifndef _LINUX
	WORD wVersionRequested = MAKEWORD(1,1);
	WSADATA data;
	int error_code=0;
	if (WSAStartup(wVersionRequested, &data) != 0)
	{
		std::cerr << "IOboard error: WSAStartup" << std::endl;
		GetLastError();
	}
	#endif

    timeval Timeout;
    Timeout.tv_sec = 2;
    Timeout.tv_usec = 0;

    struct sockaddr_in serverAddress; 
 
    socket_id = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
 
    serverAddress.sin_addr.s_addr = inet_addr(address.c_str());
    serverAddress.sin_port = htons(port); 
    serverAddress.sin_family = AF_INET;
 
    //non-blocking
	#ifdef _LINUX
	int flags = fcntl(socket_id, F_GETFL, 0);
	flags |=O_NONBLOCK;
	if(fcntl(socket_id, F_SETFL, flags)== -1)
	{
		is_connected=false;
		return false;
	}
	#else
    unsigned long iMode = 1;
    int iResult = ioctlsocket(socket_id, FIONBIO, &iMode);
    if (iResult != NO_ERROR)
    {
		is_connected=false;
		return false;
    }
	#endif 
    if(connect(socket_id,(struct sockaddr *)&serverAddress,sizeof(serverAddress))==false)
    {	
	#ifdef _LINUX

	#else
		int e = WSAGetLastError();
		std::cerr << "IOboard error: connect, " << e << std::endl;
	#endif
		is_connected=false;
        return false;
    }	
 
	#ifdef _LINUX
	flags = fcntl(socket_id, F_GETFL, 0);
	flags = flags&~O_NONBLOCK;
	if(fcntl(socket_id, F_SETFL, flags)== -1)
	{
		is_connected=false;
        close(socket_id);
		return false;
	}
	#else
    iMode = 0;
    iResult = ioctlsocket(socket_id, FIONBIO, &iMode);
    if (iResult != NO_ERROR)
    {	
		is_connected=false;
		closesocket(socket_id);
		return false;
    }
	#endif

 
    fd_set Write, Err;
    FD_ZERO(&Write);
    FD_ZERO(&Err);
    FD_SET(socket_id, &Write);
    FD_SET(socket_id, &Err);
 
    //is socket ready
    select(0,NULL,&Write,&Err,&Timeout);			
    if(FD_ISSET(socket_id, &Write)) 
    {	
		is_connected=true;
        return true;
    }
 
	is_connected=false;
	#ifdef _LINUX
        close(socket_id);
	#else
        closesocket(socket_id);
    #endif
    return false;
}

void CIOBoardTCP::SetNonBlocking(bool non_blocking)
{
	#ifdef _LINUX
	int flags = fcntl(socket_id, F_GETFL, 0);
	if (flags < 0) return;
	flags = non_blocking ? (flags|O_NONBLOCK) : (flags&~O_NONBLOCK);
	fcntl(socket_id, F_SETFL, flags);
#else
	unsigned long mode = non_blocking ? 1 : 0;
	ioctlsocket(socket_id, FIONBIO, &mode);
#endif
}

bool CIOBoardTCP::Timer(void)
{	
	if(is_connected==false)
	{
		return Connect();
	}
    else 
	{
		int num;
		unsigned char buffer[128];	
		memset(buffer,0,128);
		IO_BOARD_PACKET_HEADER header;
		if(first_read || only_monitoring)
		{
			if(!GetData())
			{
			#ifdef _LINUX
				close(socket_id);
			#else
				closesocket(socket_id);
			#endif
				is_connected=false;
				return false;
			}
		}
		else
		{
			if(!EncodeData())
			{
			#ifdef _LINUX
				close(socket_id);
			#else
				closesocket(socket_id);
			#endif
				is_connected=false;
				return false;
			}
		}

header_resync:
		num = recv(socket_id,(char*)&header,sizeof(IO_BOARD_PACKET_HEADER),0);
		if(num==sizeof(IO_BOARD_PACKET_HEADER))
		{
			if(header.magic!=PACKET_MAGIC)
				goto header_resync;
			if(header.length>0)
			{
				num = recv(socket_id,(char*)&buffer[0],header.length,0);
				if (num == -1)
				{
				#ifdef _LINUX
					close(socket_id);
				#else
					closesocket(socket_id);
				#endif
					is_connected=false;
					return false;
				}
					if(header.packet_type == PACKET_DATA)
				{
					DecodeData(buffer,header.length);
					is_data_ready=true;
				}
				else if(header.packet_type == PACKET_SETTINGS_GET)
				{
					std::unique_lock<std::mutex> lck(get_mutex);
					get_is_set = true;
					get_cv.notify_all();
					memcpy(settings_get,&buffer[0],settings_size);
				}
				else if(header.packet_type == PACKET_IDENTIFICATION)
				{
					std::unique_lock<std::mutex> lck(get_mutex);
					get_is_set = true;
					get_cv.notify_all();
					fw_version = std::string((char*)&buffer[0],12);
					hw_version = std::string((char*)&buffer[12],40);
				}
				
			}
		}
		if (num == -1)
		{
		#ifdef _LINUX
			close(socket_id);
		#else
			closesocket(socket_id);
		#endif
			is_connected=false;
			return false;
		}
	}
	return is_connected;
}

bool CIOBoardTCP::EncodeData(void)
{
	unsigned char *buffer = new unsigned char[sizeof(IO_BOARD_PACKET_HEADER)+2+NUMBER_OF_DO+2+NUMBER_OF_PWM*sizeof(std::uint16_t)];
	IO_BOARD_PACKET_HEADER *header = (IO_BOARD_PACKET_HEADER *)&buffer[0];
	header->magic = 0x11223456;
	header->packet_type = PACKET_DATA_DO;
	header->length=2+NUMBER_OF_DO+2+NUMBER_OF_PWM*sizeof(std::uint16_t);

	unsigned char *begin = &buffer[sizeof(IO_BOARD_PACKET_HEADER)];

	if(NUMBER_OF_DO)
	{
		begin[0] = DATA_BLOCK_DO;
		begin[1] = (unsigned char)(NUMBER_OF_DO);
		memcpy(&begin[2],&DO[0],NUMBER_OF_DO);

		for(unsigned i=0;i<NUMBER_OF_DO;i++)
		{
			if((DO[i]>=DO_STATE_BLINK_ONCE_100MS) && (DO[i]<=DO_STATE_BLINK_ONCE_10000MS))
				DO[i]=DO_STATE_BLINK_ONCE_WAIT;
		}

		begin +=(2+NUMBER_OF_DO);
	}
	if(NUMBER_OF_PWM)
	{
		begin[0] = DATA_BLOCK_PWM;
		begin[1] = (unsigned char)(NUMBER_OF_PWM*sizeof(std::uint16_t));
		memcpy(&begin[2],&PWM[0],NUMBER_OF_PWM*sizeof(std::uint16_t));
		begin +=(2+NUMBER_OF_PWM*sizeof(std::uint16_t)); 
	}
	int length = begin-buffer;
	int num= send(socket_id,(char*)&buffer[0],length,0);
	if (num == -1)
	{

		delete[] buffer;
		is_connected=false;
		return false;
	}
	delete[] buffer;
	return true;
}

bool CIOBoardTCP::GetData(void)
{
	IO_BOARD_PACKET_HEADER header;
	header.magic = 0x11223456;
	header.packet_type = PACKET_DATA_REQUEST;
	header.length=0;
		
	int num= send(socket_id,(char*)&header,sizeof(IO_BOARD_PACKET_HEADER),0);
	if (num == -1)
	{
		is_connected=false;
		return false;
	}
	return true;
}

void CIOBoardTCP::DecodeData(unsigned char *buffer, unsigned long size)
{
	
	std::lock_guard<std::mutex> lock(cs);
	unsigned char *p = &buffer[0];
	
	while(1)
	{
	switch(*p)
	{
	case DATA_BLOCK_DI:
		if((*(p+1)==(NUMBER_OF_DI/8)) && DI)
		{
			memcpy(&DI[0],p+2,NUMBER_OF_DI/8);
			p+=((NUMBER_OF_DI/8)+2);
		}
		break;
	case DATA_BLOCK_DO:
		if((*(p+1)==(NUMBER_OF_DO)) && DO_BACK)
		{
			memcpy(&DO_BACK[0],p+2,NUMBER_OF_DO);
			p+=((NUMBER_OF_DO)+2);
			if(first_read)
			{
				memcpy(&DO[0],&DO_BACK[0],NUMBER_OF_DO);
			}
		}
		break;
	case DATA_BLOCK_DO_REAL:
		if((*(p+1)==(NUMBER_OF_DO/8)) && DO_REAL)
		{
			memcpy(&DO_REAL[0],p+2,NUMBER_OF_DO/8);
			p+=((NUMBER_OF_DO/8)+2);
		}
		break;
	case DATA_BLOCK_AI:
		if(((*(p+1)/sizeof(std::uint16_t))==(NUMBER_OF_AI)) && AI)
		{
			memcpy(&AI[0],p+2,NUMBER_OF_AI*sizeof(std::uint16_t));
			p+=((NUMBER_OF_AI*sizeof(std::uint16_t))+2);
		}
		break;
	case DATA_BLOCK_DIP:
		DIP=*(p+2);
		p+=3;
		break;
	case DATA_BLOCK_PWM:
		if(((*(p+1)/sizeof(std::uint16_t))==(NUMBER_OF_PWM)) && PWM)
		{
			memcpy(&PWM_BACK[0],p+2,NUMBER_OF_PWM*sizeof(std::uint16_t));
			p+=((NUMBER_OF_PWM*sizeof(std::uint16_t))+2);
			if(first_read)
			{
				memcpy(&PWM[0],&PWM_BACK[0],NUMBER_OF_PWM*sizeof(std::uint16_t));
			}
		}
		break;
	default:
		p++;
		break;
	
	}
	if(*p==0 || p>&buffer[0+size])
	{
		if(first_read)
		{
			first_read=false;
		}
		break;
	}
	}

}

void CIOBoardTCP::TimerThread()
{
    while(!stop_thread)
	{
		if(Timer()==false)
			std::this_thread::sleep_for( std::chrono::microseconds(250) );
		else
			std::this_thread::sleep_for( std::chrono::microseconds(20) );
    }
}


void CIOBoardTCP::SendSettings(void *settings)
{
	if(is_connected==false)
		return;
	unsigned char *packet_array;
	switch(board_model)
	{
		default:
		case CANLAB_BOARD_MODEL_1:
			{
			packet_array = new unsigned char[sizeof(IO_BOARD_PACKET_HEADER)+sizeof(CANLAB_IO_BOARD_MODEL_1_SETTINGS)];
			CANLAB_IO_BOARD_MODEL_1_SETTINGS *s = (CANLAB_IO_BOARD_MODEL_1_SETTINGS *)&packet_array[sizeof(IO_BOARD_PACKET_HEADER)];
			*s = *((CANLAB_IO_BOARD_MODEL_1_SETTINGS *)settings);
			}
			break;
		case VRG_BOARD_MODEL_1:
			{
			packet_array = new unsigned char[sizeof(IO_BOARD_PACKET_HEADER)+sizeof(VRG_IO_BOARD_MODEL_1_SETTINGS)];
			VRG_IO_BOARD_MODEL_1_SETTINGS *s = (VRG_IO_BOARD_MODEL_1_SETTINGS *)&packet_array[sizeof(IO_BOARD_PACKET_HEADER)];
			*s = *((VRG_IO_BOARD_MODEL_1_SETTINGS *)settings);
			}
			break;
		case VRG_BOARD_MODEL_2:
			{
			packet_array = new unsigned char[sizeof(IO_BOARD_PACKET_HEADER)+sizeof(VRG_IO_BOARD_MODEL_2_SETTINGS)];
			VRG_IO_BOARD_MODEL_2_SETTINGS *s = (VRG_IO_BOARD_MODEL_2_SETTINGS *)&packet_array[sizeof(IO_BOARD_PACKET_HEADER)];
			*s = *((VRG_IO_BOARD_MODEL_2_SETTINGS *)settings);
			}
			break;
	}
	IO_BOARD_PACKET_HEADER *header = (IO_BOARD_PACKET_HEADER *)&packet_array[0];
	header->magic = 0x11223456;
	header->packet_type = PACKET_SETTINGS_SET;
	header->length=settings_size; 

	send(socket_id,(char*)&packet_array[0],sizeof(IO_BOARD_PACKET_HEADER)+settings_size,0);
}

bool CIOBoardTCP::GetSettings(void *settings)
{
	if(is_connected==false)
		return false;
	IO_BOARD_PACKET_HEADER packet;
	packet.magic = 0x11223456;
	packet.packet_type = PACKET_SETTINGS_GET;
	packet.length=0;

	{
		std::lock_guard<std::mutex> lck(get_mutex);
		get_is_set = false;
	}

	if(send(socket_id,(char*)&packet,sizeof(packet),0)==sizeof(packet))
	{
		std::unique_lock<std::mutex> lck(get_mutex);
		if (get_cv.wait_for(lck, std::chrono::milliseconds(100)) == std::cv_status::timeout)
			return false;
		if (get_is_set)
		{
			memcpy(settings,settings_get,settings_size);
		}
		return get_is_set;
	}
	return false;
}

bool CIOBoardTCP::GetIdentification(std::string &fw_ver,std::string &hw_ver)
{
	if(is_connected==false)
		return false;
	IO_BOARD_PACKET_HEADER packet;
	packet.magic = 0x11223456;
	packet.packet_type = PACKET_IDENTIFICATION;
	packet.length=0;

	{
		std::lock_guard<std::mutex> lck(get_mutex);
		get_is_set = false;
	}

	send(socket_id,(char*)&packet,sizeof(packet),0);
	{
		std::unique_lock<std::mutex> lck(get_mutex);
		if (get_cv.wait_for(lck, std::chrono::milliseconds(100)) == std::cv_status::timeout)
			return false;
		if (get_is_set)
		{
			fw_ver = fw_version;
			hw_ver = hw_version;
		}
		return get_is_set;
	}
}


void CIOBoardTCP::Calibration(unsigned int cmd,unsigned int channel)
{
	if(is_connected==false)
		return;
	if(channel>=NUMBER_OF_AI)
		return;
	IO_BOARD_PACKET_CALIBRATION packet;
	packet.header.magic = 0x11223456;
	packet.header.packet_type = PACKET_CALIBRATION_CMD;
	packet.header.length=2; 
	packet.channel=(unsigned char)channel;
	packet.point = (unsigned char)cmd;
	send(socket_id,(char*)&packet,sizeof(packet),0);
}

void CIOBoardTCP::ResetBoard(void)
{
	if(is_connected==false)
		return;
	IO_BOARD_PACKET_HEADER packet;
	packet.magic = 0x11223456;
	packet.packet_type = PACKET_RESET;
	packet.length=0; 
	send(socket_id,(char*)&packet,sizeof(packet),0);
}

