#ifndef IOBOARD_TCP_CARD_H
#define IOBOARD_TCP_CARD_H

#include <string>
#include <thread>
#include <mutex>
#include <chrono>
#include <condition_variable>
#include "CIOboard.h"
#include "CIOboardTCPDefs.h"

#ifdef _LINUX
    #include <sys/socket.h>
    #include <unistd.h>
    #include <fcntl.h>
    #include <sys/types.h>
    #include <stropts.h>
	#include <sys/ioctl.h>
	#include <netinet/in.h>
	#include <arpa/inet.h>
#else
    #include <winsock2.h>
#endif

#include <cstdint>

#pragma pack(1)
typedef struct 
{
	std::uint8_t	mac[6];
	std::uint8_t	ip[4];
    std::uint8_t	mask[4];
    std::uint8_t	gw[4];
    std::uint16_t	port;	
    
    std::uint8_t	do_init[2];		//Nastaveni digitalnich vystupu po pripojeni napajeni k desce
	std::uint16_t	pwm_init[2];
    std::uint8_t	ai_enable[2];	//Pokud nejsou zapojeny nektere analogove vstupy, je mozne je zakazat a zvetsit frekvenci mereni AN vstupu 
} CANLAB_IO_BOARD_MODEL_1_SETTINGS;

typedef struct 
{
	std::uint8_t	mac[6];
	std::uint8_t	ip[4];
    std::uint8_t	mask[4];
    std::uint8_t	gw[4];
    std::uint16_t	port;	
    
    std::uint8_t	do_init;
    std::uint8_t	ai_enable[2];
} VRG_IO_BOARD_MODEL_1_SETTINGS;

typedef struct 
{
	std::uint8_t	mac[6];
	std::uint8_t	ip[4];
    std::uint8_t	mask[4];
    std::uint8_t	gw[4];
    std::uint16_t	port;	
    
    std::uint8_t	do_init;
    std::uint16_t	pwm_init[2];
} VRG_IO_BOARD_MODEL_2_SETTINGS;


typedef struct 
{
	IO_BOARD_PACKET_HEADER header;
	std::uint8_t channel;
	std::uint8_t point;
} IO_BOARD_PACKET_CALIBRATION;

#pragma pack()


class CIOBoardTCP : public CIOBoard
{
public:	
	//Je li CreateOwnThread nastaveno na true, je vytvaren vlastni thread pro komunikaci, pokud 
	//se pouzije false, musi aplikace zabezpecit vlastni volani metory Timer.
	//Varianta bez vlastniho threadu je uzitecna pri praci s vice boardy paralelne.
	//Nastaveni OnlyMnitoring je pouzitelne v rezimu kdy se pouziva vice klientu TCP nebo TCP klienta a USB
	//V tomto rezimu klient neodesila data na vystupy aby nedochazelo k nekonzistenci dat mezi pozadavky vice klientu
	CIOBoardTCP(const unsigned BoardModel, const std::string Address, const unsigned Port, const bool CreateOwnThread = true, const bool OnlyMonitoring = false);
	virtual ~CIOBoardTCP();

	//Odeslani nastaveni do desky, nastaveni je mozne pouze prostrednictvim TCP
	void SendSettings(void* settings);

	//Precteni nastaveni z desky, precteni je mozne pouze prostrednictvim TCP
	bool GetSettings(void *settings);

	//Zjisteni verze FW a HW
	bool GetIdentification(std::string &fw_version,std::string &hw_version);
	
	//Kalibrace rozsahu AN vstupu. Nastavuji se 3 body - minimum, maximum a stred.
	//Kalibraci je nasledne nutno ulozit.
	void Calibration(unsigned int cmd, unsigned int channel=0);

	//Reset deky po zmene nastaveni.
	void ResetBoard(void);

	virtual E_IO_BOARD_CONNECTION_TYPE GetConnectionType(void) {return IO_BOARD_CON_TCP;};

	virtual bool Timer(void);
private:
	bool Connect(void);
	void SetNonBlocking(bool non_blocking);
	void TimerThread();
	bool GetData(void);
	bool EncodeData(void);
	void DecodeData(unsigned char *buffer, unsigned long size);


private:
	unsigned port;
	int socket_id; 
	std::string address; 
	bool own_thread;
    std::thread timer_thread;
    bool stop_thread;
    std::mutex get_mutex;
    std::condition_variable get_cv;
    bool get_is_set;
	unsigned char *settings_get;
	unsigned settings_size;
	std::string	 fw_version;
	std::string	 hw_version;
};

#endif