#ifndef IOBOARD_TCP_BOOTLOADER_H
#define IOBOARD_TCP_BOOTLOADER_H

#include <string>
#include <vector>
#include <thread>
#include <mutex>
#include <chrono>
#include <condition_variable>
#include "CIOBoard.h"
#include "CIOBoardTCPDefs.h"
#ifdef _LINUX
    #include <sys/socket.h>
    #include <unistd.h>
    #include <fcntl.h>
    #include <sys/types.h>
    #include <stropts.h>
	#include <sys/ioctl.h>
#else
    #include <winsock2.h>
#endif

#include <cstdint>

#pragma pack(1)

#define HEX_ROW_SIZE	64

typedef struct 
{
	IO_BOARD_PACKET_HEADER header;
	std::uint32_t hex_row_index;
	std::uint8_t hex_row[HEX_ROW_SIZE];
} IO_BOARD_PACKET_HEX_ROW;

typedef struct 
{
	IO_BOARD_PACKET_HEADER header;
	std::uint32_t hex_row_index;
} IO_BOARD_PACKET_HEX_ROW_RESPONSE;

#pragma pack()

const int BOOT_STATUS_BAD_FILE = 1;
const int BOOT_STATUS_WAIT_FOR_CONNECTION = 2;
const int BOOT_STATUS_ERASE = 4;
const int BOOT_STATUS_RUNNING = 5;
const int BOOT_STATUS_DONE= 6;
const int BOOT_STATUS_BOARD_NOT_READY= 7;

class CIOBoardTCPBootloader
{
public:	
	CIOBoardTCPBootloader(const std::string Address, const unsigned Port, std::string HexFilePath, bool CreateOwnThread = true);
	virtual ~CIOBoardTCPBootloader();

	bool IsConnected(void) {return is_connected;};
	int GetBootStatus(unsigned &progress) {progress = next_hex_row; return status;};

private:
	bool LoadHexFile(std::string filepath);
	bool Connect(void);
	void SetNonBlocking(bool non_blocking);
	bool Timer(void);
	void TimerThread();
	bool SendRow(void);

private:
	std::vector<std::string> hex;
	int next_hex_row;
	unsigned port;
	int socket_id; 
	std::string address; 
	bool own_thread;
    std::thread timer_thread;
    bool stop_thread;
	bool is_connected;
	int  status;


};

#endif