#include "CIOboardUSB.h"
#include <string.h>
typedef enum
{
    COMMAND_GET_IO_STATUS = 0x81,
    COMMAND_SET_IO = 0x82
} CUSTOM_HID_COMMANDS;

CIOBoardUSB::CIOBoardUSB(const unsigned BoardModel, const unsigned short VID, const unsigned short PID, char *SerialNumber, const bool CreateOwnThread, const bool OnlyMonitoring)
	: CIOBoard(OnlyMonitoring, BoardModel)
	, vendor_id(VID)
	, product_id(PID)
	, stop_thread(false)
	, use_serial_number(SerialNumber)
	, own_thread(CreateOwnThread)
	, last_error_code(NULL)
{
	hid_init();

	//hid_device_info *d = hid_enumerate(0x04d8, 0x003f);

	if(use_serial_number)
		serial_number = std::wstring(SerialNumber, SerialNumber+strlen(SerialNumber));
    memset((void*)&buffer[0], 0x00, sizeof(buffer));

	if(own_thread)
		timer_thread = std::thread(&CIOBoardUSB::TimerThread,this);
}

CIOBoardUSB::~CIOBoardUSB()
{
	if(own_thread)
	{
		stop_thread = true;
		if(timer_thread.joinable()) 
			timer_thread.join();
	}
	if(is_connected)
		hid_close(device);
	hid_exit();
}

bool CIOBoardUSB::Timer(void)
{	
	if(is_connected==false)
	{
		if(use_serial_number==NULL)
		{
			device = hid_open(0x04d8, 0x003f, NULL);
		}
		else
		{
			device = hid_open(0x04d8, 0x003f, (wchar_t*)serial_number.c_str());
		}
        if (device) 
		{
            is_connected = true;
            hid_set_nonblocking(device, true);
			return true;
        }
		else return false;
	}
    else 
	{
		//periodicky odesilam stav vystupu
		//deska na zpravu se stavem odpovida vsech vstupu a vystupu 
		if(first_read || only_monitoring)
		{
			GetData();
		}
		else
		{
			EncodeData();
		}
		if (hid_write(device, buffer, sizeof(buffer)) == -1)
        {
			last_error_code = hid_error(device);
			hid_close(device);
			device = NULL;
			is_connected = false;
            return false;
        }

        if(hid_read(device, buffer, sizeof(buffer)) == -1)
        {
			last_error_code = hid_error(device);
			hid_close(device);
			device = NULL;
		    is_connected = false;
            return false;
        }
        
		if(buffer[0] == COMMAND_GET_IO_STATUS) 
		{
			DecodeData();
			is_data_ready=true;
        }
	}
	return true;
}

void CIOBoardUSB::EncodeData(void)
{
	std::lock_guard<std::mutex> lock(cs);
	memset((void*)&buffer[0], 0x00, sizeof(buffer));
	buffer[0] = 0x00;
	buffer[1] = COMMAND_SET_IO;

	buffer[2] = 0;
	if(NUMBER_OF_DO)
		buffer[2]+=	(unsigned char)(2+NUMBER_OF_DO);
	if(NUMBER_OF_PWM)
		buffer[2]+= (unsigned char)(2+(NUMBER_OF_PWM*sizeof(std::uint16_t))); //bez hlavicky
	/*
	buffer[3] = DATA_BLOCK_DO;
	buffer[4] = NUMBER_OF_DO;
	memcpy(&buffer[5],&DO[0],NUMBER_OF_DO);

	buffer[5+NUMBER_OF_DO] = DATA_BLOCK_PWM;
	buffer[5+NUMBER_OF_DO+1] = NUMBER_OF_PWM;
	memcpy(&buffer[5+NUMBER_OF_DO+2],&PWM[0],NUMBER_OF_PWM*sizeof(unsigned __int16));
	*/
	unsigned char *begin = &buffer[3];

	if(NUMBER_OF_DO)
	{
		begin[0] = DATA_BLOCK_DO;
		begin[1] = (unsigned char)(NUMBER_OF_DO);
		memcpy(&begin[2],&DO[0],NUMBER_OF_DO);
		begin +=(2+NUMBER_OF_DO); 
	}

	if(NUMBER_OF_PWM)
	{
		begin[0] = DATA_BLOCK_PWM;
		begin[1] = (unsigned char)(NUMBER_OF_PWM*sizeof(std::uint16_t));
		memcpy(&begin[2],&PWM[0],NUMBER_OF_PWM*sizeof(std::uint16_t));
		begin +=(2+NUMBER_OF_PWM*sizeof(std::uint16_t)); 
	}
}

void CIOBoardUSB::GetData(void)
{
	memset((void*)&buffer[0], 0x00, sizeof(buffer));
	buffer[0] = 0x00;
	buffer[1] = COMMAND_GET_IO_STATUS;
}

void CIOBoardUSB::DecodeData(void)
{
	std::lock_guard<std::mutex> lock(cs);
	unsigned char *p = &buffer[2];
	unsigned char size = buffer[1];
	while(1)
	{
	switch(*p)
	{
	case DATA_BLOCK_DI:
		if((*(p+1)==(NUMBER_OF_DI/8)) && DI)
		{
			memcpy(&DI[0],p+2,NUMBER_OF_DI/8);
			p+=((NUMBER_OF_DI/8)+2);
		}
		else
		{
			p+=*(p+1);
		}
		break;
	case DATA_BLOCK_DO:
		if((*(p+1)==(NUMBER_OF_DO)) && DO_BACK)
		{
			memcpy(&DO_BACK[0],p+2,NUMBER_OF_DO);
			p+=((NUMBER_OF_DO)+2);
			if(first_read)
			{
				first_read=false;
				memcpy(&DO[0],&DO_BACK[0],NUMBER_OF_DO);
			}
		}
		else
		{
			p+=*(p+1);
		}
		break;
	case DATA_BLOCK_DO_REAL:
		if((*(p+1)==(NUMBER_OF_DO/8)) && DO_REAL)
		{
			memcpy(&DO_REAL[0],p+2,NUMBER_OF_DO/8);
			p+=((NUMBER_OF_DO/8)+2);
		}
		else
		{
			p+=*(p+1);
		}
		break;
	case DATA_BLOCK_AI:
		if(((*(p+1)/sizeof(std::uint16_t))==(NUMBER_OF_AI)) && AI)
		{
			memcpy(&AI[0],p+2,NUMBER_OF_AI*sizeof(std::uint16_t));
			p+=((NUMBER_OF_AI*sizeof(std::uint16_t))+2);
		}
		else
		{
			p+=*(p+1);
		}
		break;
	case DATA_BLOCK_DIP:
		DIP=*(p+2);
		p+=3;
		break;
	case DATA_BLOCK_PWM:
		if(((*(p+1)/sizeof(std::uint16_t))==(NUMBER_OF_PWM)) && PWM)
		{
			memcpy(&PWM_BACK[0],p+2,NUMBER_OF_PWM*sizeof(std::uint16_t));
			p+=((NUMBER_OF_PWM*sizeof(std::uint16_t))+2);
			if(first_read)
			{
				memcpy(&PWM[0],&PWM_BACK[0],NUMBER_OF_PWM*sizeof(std::uint16_t));
			}
		}
		else
		{
			p+=*(p+1);
		}
		break;
	default:
		p++;
		break;	
	}
	if(*p==0 || p>&buffer[1+size])
		break;
	}
}

void CIOBoardUSB::TimerThread()
{
    while(!stop_thread)
	{
		if(Timer()==false)
			std::this_thread::sleep_for( std::chrono::microseconds(250) );
		else
			std::this_thread::sleep_for( std::chrono::microseconds(20) );
    }
}
