#ifndef IOBOARD_USB_CARD_H
#define IOBOARD_USB_CARD_H

#include <string>
#include <thread>
#include <mutex>
#include <chrono>
#include "hidapi.h"
#include "CIOboard.h"

const unsigned short IO_BOARD_VID = 0x04d8;
const unsigned short IO_BOARD_PID = 0x003f;

class CIOBoardUSB : public CIOBoard
{
public:
	//Je li pripojen jen jeden IO board, lze zadat parametr SerialNumber = NULL
	//V pripade ze je treba rozlisit vice pripojenych zarizeni, zadava se polozka SerialNumber.
	//Ta se nastavuje podle hodnoty DIP prepinace.
	//Pokud z je na DIP 1-5 nastavena napriklad binarni kombiace cisla 6, v Zarizeni a tiskarny se zobrazuje
	//text: CANLAB IO board USB/ETH 06
	//Pro otevreni tohoto bordu zadejte jako SerialNumber retezec "06"
	CIOBoardUSB(const unsigned BoardModel, const unsigned short VID, const unsigned short PID, char *SerialNumber, const bool CreateOwnThread = true, const bool OnlyMonitoring = false);
	virtual ~CIOBoardUSB();
	
	E_IO_BOARD_CONNECTION_TYPE GetConnectionType(void) {return IO_BOARD_CON_USB;};

	virtual bool Timer(void);
private:
	void TimerThread();
	void GetData(void);
	void EncodeData(void);
	void DecodeData(void);

private:
	unsigned short vendor_id;
	unsigned short product_id;
	std::wstring serial_number;
	char *use_serial_number;
	hid_device *device;
	unsigned char buffer[65];

	bool own_thread;
    std::thread timer_thread;
    bool stop_thread;
	const wchar_t* last_error_code;

};

#endif